IMPLEMENTATION MODULE Client;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Types, Lib2, Lib3, SO, Crc, MD5, Tmrs, Tmr, UFTP, Config, SrvIO, File, Cfg, Log;

CONST
  mid = Cfg.midClient;

  period =    1;  (* ms, update period *)
  toSes  = 2000;  (* ms, session request period *)
  toDat  =  200;  (* ms, data request period *)
  fglMax = UFTP.lenMaxFgm;  (* adjust here *)
  wndMax = 32;

  lenBuf = fglMax;

TYPE
  TState = ( stOpen, stData, stClose );

  TSesDesc = RECORD
    t: Tmr.TTicks;    (* session request *)
    f: UFTP.TFgmLen;
    l: UFTP.TFileLen;
    d: MD5.TDgstDesc;
    r: UFTP.TReference;
    n: UFTP.TFgmIdx;  (* fgm cnt *)
    e: UFTP.TFgmLen;  (* last fgm *)
  END;

  TWndIdx = UFTP.TFgmIdx[ 0 .. wndMax - 1 ];
  TWndSet = SET OF TWndIdx;

  TWndDesc = RECORD
    i: UFTP.TFgmIdx;  (* lower window boundary *)
    j: UFTP.TFgmIdx;  (* upper window boundary + 1;  j - i <= wndMax *)
    w: UFTP.TFgmIdx[ 0 .. wndMax ];  (* windows size *)
    r: TWndSet;       (* requested fragments *)
    d: TWndSet;       (* done fragments *)
    t: ARRAY TWndIdx OF Tmr.TTicks;  (* request times *)
  END;

(*
  ses.n fragments
  fgmt state: fsNull, fsRequested, fsDone
  window: 0 <= wnd.i <= wnd.j <= ses.n;
*)

  TDataDesc = RECORD
    st : TState;
    ers: TErrCnts;
    ses: TSesDesc;
    wnd: TWndDesc;
  END;

VAR
  data: TDataDesc;
  tmr: Tmrs.TTimer;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE ShowErrCnts();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
  e: TErr;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'ShowErrCnts' );
  FOR e := MIN(TErr) TO MAX(TErr) DO
    SO.WrChar( b, ' ' );
    SO.WrLngHex( b, data.ers[e] MOD 1000H, 3 );
  END;
  Message( s );
END ShowErrCnts;


PROCEDURE WrWnd( VAR b: SO.TBuffer );
VAR
  i: UFTP.TFgmIdx;
  k: TWndIdx;
BEGIN
  SO.WrStr( b, ' wnd ' );
  SO.WrLngCard( b, data.wnd.i, 1 );
  SO.WrChar( b, ' ' );
  SO.WrLngCard( b, data.wnd.j, 1 );
  SO.WrStr( b, ' req ' );
  i := data.wnd.i;
  WHILE i < data.wnd.j DO
    k := i MOD wndMax;
    IF k IN data.wnd.r THEN
      SO.WrLngCard( b, i, 1 );
      SO.WrChar( b, ' ' );
    END;
    INC( i );
  END;
  SO.WrStr( b, 'don ' );
  i := data.wnd.i;
  WHILE i < data.wnd.j DO
    k := i MOD wndMax;
    IF k IN data.wnd.d THEN
      SO.WrLngCard( b, i, 1 );
      SO.WrChar( b, ' ' );
    END;
    INC( i );
  END;
END WrWnd;


PROCEDURE IncErrCnt( e: TErr );
BEGIN
  Lib2.IncL( data.ers[e], 1 );
  Lib2.IncL( data.ers[eNull],1  );
END IncErrCnt;

PROCEDURE OnRx( dat-: ARRAY OF BYTE; len: CARDINAL );
VAR
  tel: POINTER TO UFTP.TTelDesc;

  PROCEDURE SesRp();
  CONST
    lenMin = SIZE( UFTP.TSesRpDesc );
    lenMax = lenMin;
  VAR
    f: UFTP.TFgmLen;
    l: UFTP.TFileLen;

    PROCEDURE Show();
    CONST
      ena = TRUE;
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..127] OF CHAR;
      m: MD5.TDgstStr;
    BEGIN
      IF ~ ena THEN RETURN END;
      MD5.MkStr( m, data.ses.d );
      SO.Init( b, s );
      SO.WrStr( b, 'SesRp: fgl ' );
      SO.WrLngCard( b, data.ses.f, 1 );
      SO.WrStr( b, ' len ' );
      SO.WrLngCard( b, data.ses.l, 1 );
      SO.WrStr( b, ' md5 ' );
      SO.WrStr( b, m );
      SO.WrStr( b, ' ref ' );
      SO.WrLngCard( b, data.ses.r, 1 );
      SO.WrStr( b, ' cnt ' );
      SO.WrLngCard( b, data.ses.n, 1 );
      SO.WrStr( b, ' lst ' );
      SO.WrLngCard( b, data.ses.e, 1 );
      Message( s );
    END Show;

  BEGIN
    IF data.st # stOpen THEN RETURN END;
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eSesRpLen );
      RETURN;
    END;
    CASE tel^.srp.e OF
    | UFTP.errNull: (* ok *)
    | UFTP.errRID : Fatal( 'File not found.' );
    | UFTP.errRes : Fatal( 'Server full.' );
    ELSE
      Fatal( 'Unknown error.' );
    END;
    f := tel^.srp.f;
    l := tel^.srp.l;
    IF ( f = 0 ) OR ( f > fglMax ) THEN
      IncErrCnt( eSesRpFgl );
      RETURN;
    END;
    (*
    data.ses.t := Tmr.Ticks();
    *)
    data.ses.f := f;
    data.ses.l := l;
    data.ses.d := tel^.srp.d;
    data.ses.r := tel^.srp.r;
    data.ses.n := l DIV f + ORD( l MOD f > 0 );
    data.ses.e := l MOD f;
    Lib2.Fill( data.wnd, 0 );
    data.wnd.w := wndMax;  (* adjust bandwidth here, period may be used too *)
    File.Open( 'out.bin', f, l, data.ses.d );
    IF data.ses.n > 0 THEN data.st := stData ELSE data.st := stClose END;
    Show();
  END SesRp;

  PROCEDURE DatRp();
  CONST
    lenMin = UFTP.lenMinDatRpDesc;
    lenMax = UFTP.lenMaxDatRpDesc;
  VAR
    i, m: UFTP.TFgmIdx;
    k: TWndIdx;

    PROCEDURE Show();
    CONST
      ena = TRUE;
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..511] OF CHAR;
    BEGIN
      IF ~ ena THEN RETURN END;
      SO.Init( b, s );
      SO.WrStr( b, 'DatRp: idx ' );
      SO.WrLngCard( b, i, 1 );
      SO.WrStr( b, ' len ' );
      SO.WrLngCard( b, len, 1 );
      WrWnd( b );
      Message( s );
    END Show;

  BEGIN
    IF data.st # stData THEN RETURN END;
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eDatRpLen );
      RETURN;
    END;
    DEC( len, SIZE( tel^.drp.i ) );
    i := tel^.drp.i;
    IF ( i < data.wnd.i ) OR ( i >= data.wnd.j ) THEN
      IncErrCnt( eDatRpWnd );  (* not really an error, just for debugging *)
      RETURN;
    END;
    m := data.ses.n - 1;  (* max fgm idx *)
    IF ( i < m ) & ( len # data.ses.f ) OR
       ( i = m ) & ( len # data.ses.e ) THEN
      IncErrCnt( eDatRpFgl );
      RETURN;
    END;
    k := i MOD wndMax;
    INCL( data.wnd.d, k );
    IF k IN data.wnd.r THEN
      File.Write( i, tel^.drp.d, len );
    ELSE
      IncErrCnt( eDatRpReq );  (* not really an error, just for debugging *)
    END;
    Show();
  END DatRp;

CONST
  lenMin = UFTP.lenFrmOvr;
  lenMax = UFTP.lenMaxTel;
BEGIN
  IF ( len < lenMin ) OR ( len > lenMax ) THEN IncErrCnt( eLen ); RETURN END;
  IF ~ Crc.Ok( dat, len ) THEN IncErrCnt( eCrc ); RETURN END;
  tel := ADR( dat );
  DEC( len, lenMin );
  IF tel^.hdr.m # UFTP.mag THEN IncErrCnt( eMag ); RETURN END;
  IF tel^.hdr.v # UFTP.ver THEN IncErrCnt( eVer ); RETURN END;
  CASE tel^.hdr.c OF
  | UFTP.cmdSesRp: SesRp();
  | UFTP.cmdDatRp: DatRp();
  ELSE
    IncErrCnt( eCmd );
  END;
END OnRx;


PROCEDURE DoOpen();
VAR
  r: RECORD h: UFTP.THdrDesc; b: UFTP.TSesRqDesc; t: UFTP.TTrlDesc END;

  PROCEDURE ShowTel();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'SesRq: fgl ' );
    SO.WrCard( b, r.b.f, 1 );
    SO.WrStr( b, ' rid ' );
    SO.WrLngHex( b, LONGCARD( r.b.r ), 1 );
    Message( s );
  END ShowTel;

VAR
  t: Tmr.TTicks;
BEGIN
  t := Tmr.Ticks();
  IF Tmr.Diff( data.ses.t, t ) < toSes THEN RETURN END;
  data.ses.t := t;
  Lib2.Fill( r, 0 );
  r.h.m := UFTP.mag;
  r.h.v := UFTP.ver;
  r.h.c := UFTP.cmdSesRq;
  r.b.f := fglMax;
  r.b.r := 2;
  r.t := Crc.Swap( Crc.UpdL( Crc.crcInit, r, SIZE(r) - SIZE(r.t) ) );
  SrvIO.Send( r, SIZE( r ) );
  ShowTel();
END DoOpen;

PROCEDURE DoData();
VAR
  tel: RECORD h: UFTP.THdrDesc; b: UFTP.TDatRqDesc; t: UFTP.TTrlDesc END;

  PROCEDURE ShowTel();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'DatRq: idx ' );
    SO.WrLngCard( b, tel.b.i, 1 );
    SO.WrStr( b, ' set ' );
    SO.WrLngHex( b, LONGCARD( tel.b.s ), 8 );
    Message( s );
  END ShowTel;

  PROCEDURE Flush();
  BEGIN
    IF ( tel.h.m = 0 ) OR ( tel.b.s = UFTP.TFgmSet{} ) THEN RETURN END;
    tel.t := Crc.Swap( Crc.UpdL( Crc.crcInit, tel, SIZE(tel) - SIZE(tel.t) ) );
    SrvIO.Send( tel, SIZE( tel ) );
    ShowTel();
  END Flush;

  PROCEDURE Request( i: UFTP.TFgmIdx );
  BEGIN
    IF tel.h.m = 0 THEN
      tel.h.m := UFTP.mag;
      tel.h.v := UFTP.ver;
      tel.h.c := UFTP.cmdDatRq;
      tel.b.r := data.ses.r;
      tel.b.s := UFTP.TFgmSet{};
    END;
    IF tel.b.s = UFTP.TFgmSet{} THEN tel.b.i := i END;
    IF i - tel.b.i > MAX( UFTP.TRelFgmIdx ) THEN
      Flush();
      tel.b.i := i;
      tel.b.s := UFTP.TFgmSet{};
    END;
    INCL( tel.b.s, i - tel.b.i );
  END Request;

  PROCEDURE ShowWnd( n: CARDINAL );
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..511] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'DoData' );
    SO.WrCard( b, n, 1 );
    SO.WrChar( b, ':' );
    WrWnd( b );
    Message( s );
  END ShowWnd;

  PROCEDURE ShowTO( i: UFTP.TFgmIdx; k: TWndIdx );
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..511] OF CHAR;
  BEGIN
    IF ~ ( k IN data.wnd.r ) THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'DoData.TO: idx ' );
    SO.WrLngCard( b, i, 1 );
    Message( s );
  END ShowTO;

VAR
  t: Tmr.TTicks;
  i: UFTP.TFgmIdx;
  k: TWndIdx;
BEGIN
  tel.h.m := 0;
  t := Tmr.Ticks();
  ShowWnd( 1 );
  (* progress wnd.i *)
  i := data.wnd.i;
  k := i MOD wndMax;
  WHILE ( i < data.wnd.j ) & ( k IN data.wnd.d ) DO
    EXCL( data.wnd.r, k );
    EXCL( data.wnd.d, k );
    INC( i );
    k := i MOD wndMax;
  END;
  IF i >= data.ses.n THEN
    data.st := stClose;
    RETURN;
  END;
  data.wnd.i := i;
  (* adjust wnd.w (not yet used) *)
  (* update wnd.j (never decreased) *)
  data.wnd.j := i + data.wnd.w;
  IF data.wnd.j > data.ses.n THEN
    data.wnd.j := data.ses.n;
  END;
  ShowWnd( 2 );
  WHILE i < data.wnd.j DO
    k := i MOD wndMax;
    IF ~ ( k IN data.wnd.d ) & ( ~ ( k IN data.wnd.r ) OR ( Tmr.Diff( data.wnd.t[k], t ) >= toDat ) ) THEN
      ShowTO( i, k );
      Request( i );
      INCL( data.wnd.r, k );
      data.wnd.t[ k ] := t;
    END;
    INC( i );
  END;
  Flush();
  ShowWnd( 3 );
END DoData;

PROCEDURE DoClose();
VAR
  r: RECORD h: UFTP.THdrDesc; b: UFTP.TCloseDesc; t: UFTP.TTrlDesc END;

  PROCEDURE ShowTel();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Close: ref ' );
    SO.WrCard( b, r.b.r, 1 );
    Message( s );
  END ShowTel;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    x: LONGCARD;  (* bandwidth in B/ms = 1000 b/s *)
  BEGIN
    IF ~ ena THEN RETURN END;
    x := data.ses.l DIV ( Tmr.Ticks() - data.ses.t );
    SO.Init( b, s );
    SO.WrStr( b, 'DoClose: ' );
    SO.WrLngCard( b, x, 1 );
    SO.WrStr( b, ' bytes/ms' );
    Message( s );
  END Show;

BEGIN
  File.Close();

  Lib2.Fill( r, 0 );
  r.h.m := UFTP.mag;
  r.h.v := UFTP.ver;
  r.h.c := UFTP.cmdClose;
  r.b.r := data.ses.r;
  r.t := Crc.Swap( Crc.UpdL( Crc.crcInit, r, SIZE(r) - SIZE(r.t) ) );
  SrvIO.Send( r, SIZE( r ) );
  ShowTel();

  Show();
  Lib3.SetReturnCode( 0 );
  HALT;
END DoClose;

PROCEDURE OnTmr( k: Tmrs.TTicks; r: Tmrs.TRef );
BEGIN
  CASE data.st OF
  | stOpen : DoOpen();
  | stData : DoData();
  | stClose: DoClose();
  ELSE
  END;
END OnTmr;


PROCEDURE Run();
BEGIN
  data.ses.t := Tmr.Sub( Tmr.Ticks(), toSes );  (* force initial open timeout *)
  SrvIO.Execute();
END Run;

PROCEDURE GetErrCnts( VAR e: TErrCnts );
BEGIN
  e := data.ers;
END GetErrCnts;


VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
BEGIN
  ShowErrCnts();
  Message( 'DoneModule' );
  oldTerm();
END DoneModule;

PROCEDURE InitModule();
BEGIN
  Log.Register( mid, 'Client' );
  Lib2.Fill( data, 0 );

  tmr := Tmrs.New();
  Tmrs.Set( tmr, 1, period );
  Tmrs.SetHnd( tmr, OnTmr, 0 );

  SrvIO.Install( OnRx );

  Lib3.Terminate( DoneModule, oldTerm );
END InitModule;

BEGIN
  InitModule();
END Client.
